"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.WinixPurifierAccessory = void 0;
const winix_api_1 = require("winix-api");
const errors_1 = require("./errors");
const DEFAULT_FILTER_LIFE_REPLACEMENT_PERCENTAGE = 10;
const DEFAULT_CACHE_INTERVAL_SECONDS = 60;
const MIN_AMBIENT_LIGHT = 0.0001;
class WinixPurifierAccessory {
    constructor(log, platform, config, accessory, override) {
        this.log = log;
        this.platform = platform;
        this.config = config;
        this.accessory = accessory;
        this.override = override;
        const { deviceId, deviceAlias } = accessory.context.device;
        this.deviceId = deviceId;
        this.latestStatus = {};
        this.cacheIntervalMs = (config.cacheIntervalSeconds ?? DEFAULT_CACHE_INTERVAL_SECONDS) * 1000;
        this.lastWinixPoll = -1;
        this.servicesInUse = new Set();
        const deviceSerial = override?.serialNumber ?? 'WNXAI00000000';
        const deviceName = override?.nameDevice ?? deviceAlias;
        const airQualityName = override?.nameAirQuality ?? 'Air Quality';
        const plasmawaveName = override?.namePlasmawave ?? 'Plasmawave';
        const ambientLightName = override?.nameAmbientLight ?? 'Ambient Light';
        const autoSwitchName = override?.nameAutoSwitch ?? 'Auto Mode';
        // Create services
        this.purifier = accessory.getService(this.platform.Service.AirPurifier) ||
            accessory.addService(this.platform.Service.AirPurifier);
        this.purifier.updateCharacteristic(this.platform.Characteristic.Name, deviceName);
        this.purifier.updateCharacteristic(this.platform.Characteristic.ConfiguredName, deviceName);
        this.servicesInUse.add(this.purifier);
        // TODO: Add handler for get/set ConfiguredName
        this.purifier.getCharacteristic(this.platform.Characteristic.Active)
            .onGet(this.getActiveState.bind(this))
            .onSet(this.setActiveState.bind(this));
        this.purifier.getCharacteristic(this.platform.Characteristic.CurrentAirPurifierState)
            .onGet(this.getCurrentState.bind(this));
        this.purifier.getCharacteristic(this.platform.Characteristic.TargetAirPurifierState)
            .onGet(this.getTargetState.bind(this))
            .onSet(this.setTargetState.bind(this));
        this.purifier.getCharacteristic(this.platform.Characteristic.RotationSpeed)
            .onGet(this.getRotationSpeed.bind(this))
            .onSet(this.setRotationSpeed.bind(this));
        this.purifier.getCharacteristic(this.platform.Characteristic.FilterLifeLevel)
            .onGet(this.getFilterLifeLevel.bind(this));
        this.purifier.getCharacteristic(this.platform.Characteristic.FilterChangeIndication)
            .onGet(this.getFilterChangeIndication.bind(this));
        this.purifierInfo = accessory.getService(this.platform.Service.AccessoryInformation) ||
            accessory.addService(this.platform.Service.AccessoryInformation);
        this.purifierInfo.updateCharacteristic(this.platform.Characteristic.Manufacturer, 'Winix');
        this.purifierInfo.updateCharacteristic(this.platform.Characteristic.SerialNumber, deviceSerial);
        this.purifierInfo.getCharacteristic(this.platform.Characteristic.FirmwareRevision)
            .onGet(() => accessory.context.device.mcuVer);
        this.purifierInfo.getCharacteristic(this.platform.Characteristic.Model)
            .onGet(() => accessory.context.device.modelName);
        this.servicesInUse.add(this.purifierInfo);
        if (config.exposeAirQuality) {
            this.airQuality = accessory.getServiceById(this.platform.Service.AirQualitySensor, 'air-quality-sensor') ||
                accessory.addService(this.platform.Service.AirQualitySensor, airQualityName, 'air-quality-sensor');
            this.airQuality.setCharacteristic(this.platform.Characteristic.Name, airQualityName);
            // TODO: Add handler for get/set ConfiguredName
            this.airQuality.setCharacteristic(this.platform.Characteristic.ConfiguredName, airQualityName);
            this.airQuality.getCharacteristic(this.platform.Characteristic.AirQuality)
                .onGet(this.getAirQuality.bind(this));
            this.servicesInUse.add(this.airQuality);
        }
        if (config.exposePlasmawave) {
            this.plasmawave = accessory.getServiceById(this.platform.Service.Switch, 'switch-plasmawave') ||
                accessory.addService(this.platform.Service.Switch, plasmawaveName, 'switch-plasmawave');
            this.plasmawave.setCharacteristic(this.platform.Characteristic.Name, plasmawaveName);
            // TODO: Add handler for get/set ConfiguredName
            this.plasmawave.setCharacteristic(this.platform.Characteristic.ConfiguredName, plasmawaveName);
            this.plasmawave.getCharacteristic(this.platform.Characteristic.On)
                .onGet(this.getPlasmawave.bind(this))
                .onSet(this.setPlasmawave.bind(this));
            this.servicesInUse.add(this.plasmawave);
        }
        if (config.exposeAmbientLight) {
            this.ambientLight = accessory.getServiceById(this.platform.Service.LightSensor, 'light-sensor-ambient') ||
                accessory.addService(this.platform.Service.LightSensor, ambientLightName, 'light-sensor-ambient');
            this.ambientLight.setCharacteristic(this.platform.Characteristic.Name, ambientLightName);
            // TODO: Add handler for get/set ConfiguredName
            this.ambientLight.setCharacteristic(this.platform.Characteristic.ConfiguredName, ambientLightName);
            this.ambientLight.getCharacteristic(this.platform.Characteristic.CurrentAmbientLightLevel)
                .onGet(this.getAmbientLight.bind(this));
            this.servicesInUse.add(this.ambientLight);
        }
        if (config.exposeAutoSwitch) {
            this.autoSwitch = accessory.getServiceById(this.platform.Service.Switch, 'switch-auto') ||
                accessory.addService(this.platform.Service.Switch, autoSwitchName, 'switch-auto');
            this.autoSwitch.setCharacteristic(this.platform.Characteristic.Name, autoSwitchName);
            // TODO: Add handler for get/set ConfiguredName
            this.autoSwitch.setCharacteristic(this.platform.Characteristic.ConfiguredName, autoSwitchName);
            this.autoSwitch.getCharacteristic(this.platform.Characteristic.On)
                .onGet(this.getAutoSwitchState.bind(this))
                .onSet(this.setAutoSwitchState.bind(this));
            this.servicesInUse.add(this.autoSwitch);
        }
        this.pruneUnusedServices();
    }
    pruneUnusedServices() {
        this.accessory.services.forEach((service) => {
            if (this.servicesInUse.has(service)) {
                return;
            }
            this.debug('pruning unused service:', service.displayName);
            this.accessory.removeService(service);
        });
    }
    getFilterLifeLevel() {
        const { filterMaxPeriod, filterReplaceDate } = this.accessory.context.device;
        const period = parseInt(filterMaxPeriod, 10);
        if (isNaN(period)) {
            this.debug('getFilterLifeLevel(): device.filterMaxPeriod is not a number:', period);
            // just assuming 100% life if filterMaxPeriod is not valid
            return 100;
        }
        // Ensure the date is in ISO 8601 format
        const isoDate = filterReplaceDate.replace(' ', 'T') + 'Z';
        const replaceDate = new Date(isoDate);
        const currentDate = new Date();
        // Total lifespan in milliseconds
        const totalLifespan = new Date(replaceDate);
        totalLifespan.setMonth(totalLifespan.getMonth() + period);
        const elapsedTime = currentDate.getTime() - replaceDate.getTime();
        const totalLifespanTime = totalLifespan.getTime() - replaceDate.getTime();
        const lifeUsed = Math.min(Math.max(elapsedTime / totalLifespanTime, 0), 1);
        const lifeLevel = Math.round((1 - lifeUsed) * 100);
        this.debug('getFilterLifeLevel()', lifeLevel);
        return lifeLevel;
    }
    getFilterChangeIndication() {
        const filterLife = this.getFilterLifeLevel();
        const replacementPercentage = this.config.filterReplacementIndicatorPercentage ?? DEFAULT_FILTER_LIFE_REPLACEMENT_PERCENTAGE;
        const shouldReplaceFilter = filterLife <= replacementPercentage ?
            this.platform.Characteristic.FilterChangeIndication.CHANGE_FILTER :
            this.platform.Characteristic.FilterChangeIndication.FILTER_OK;
        this.debug('getFilterChangeIndication() filterLife:', filterLife, 'replacementPercentage:', replacementPercentage, 'shouldReplaceFilter:', shouldReplaceFilter);
        return shouldReplaceFilter;
    }
    async getActiveState() {
        if (this.shouldUseCachedValue(this.latestStatus.power)) {
            this.debug('getActiveState() (cached)', this.latestStatus.power);
            return this.toActiveState(this.latestStatus.power);
        }
        let power;
        try {
            power = await winix_api_1.WinixAPI.getPower(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting active state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.power = power;
        this.polledWinix();
        this.debug('getActiveState()', power);
        return this.toActiveState(power);
    }
    async setActiveState(state) {
        const power = state === this.platform.Characteristic.Active.ACTIVE ? winix_api_1.Power.On : winix_api_1.Power.Off;
        this.debug(`setActiveState(${state})`, power);
        if (this.latestStatus.power === power) {
            this.debug('ignoring duplicate state set: active');
            return;
        }
        try {
            await winix_api_1.WinixAPI.setPower(this.deviceId, power);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error setting active state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.power = power;
        this.sendHomekitUpdate();
    }
    async getCurrentState() {
        if (this.shouldUseCachedValue(this.latestStatus.power)) {
            this.debug('getCurrentState() (cached)', this.latestStatus.power);
            return this.toCurrentState(this.latestStatus.power);
        }
        let power;
        try {
            power = await winix_api_1.WinixAPI.getPower(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting current state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.power = power;
        this.polledWinix();
        this.debug('getCurrentState()', power);
        return this.toCurrentState(power);
    }
    async getAutoSwitchState() {
        const targetState = await this.getTargetState();
        this.debug('getAutoSwitchState() targetState', targetState);
        // Translate target state (auto/manual mode) to auto switch state
        const result = targetState === this.platform.Characteristic.TargetAirPurifierState.AUTO ?
            this.platform.Characteristic.Active.ACTIVE : this.platform.Characteristic.Active.INACTIVE;
        this.debug('getAutoSwitchState() result', result);
        return result;
    }
    async setAutoSwitchState(state) {
        // Translate auto switch state to target state (auto/manual mode)
        const proxyState = state ?
            this.platform.Characteristic.TargetAirPurifierState.AUTO :
            this.platform.Characteristic.TargetAirPurifierState.MANUAL;
        this.debug(`setAutoSwitchState(${state}) proxyState`, proxyState);
        return this.setTargetState(proxyState);
    }
    async getTargetState() {
        if (this.shouldUseCachedValue(this.latestStatus.mode)) {
            this.debug('getTargetState() (cached)', this.latestStatus.mode);
            return this.toTargetState(this.latestStatus.mode);
        }
        let mode;
        try {
            mode = await winix_api_1.WinixAPI.getMode(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting target state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.mode = mode;
        this.polledWinix();
        this.debug('getTargetState()', mode);
        return this.toTargetState(mode);
    }
    async setTargetState(state) {
        const mode = state === this.platform.Characteristic.TargetAirPurifierState.AUTO ? winix_api_1.Mode.Auto : winix_api_1.Mode.Manual;
        this.debug(`setTargetState(${state})`, mode);
        // Don't try to set the mode if we're already in this mode
        // Fixes issues with this being set right around the time of power on
        if (this.latestStatus.mode === mode) {
            this.debug('ignoring duplicate state set: target');
            return;
        }
        try {
            await winix_api_1.WinixAPI.setMode(this.deviceId, mode);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error setting target state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.mode = mode;
        this.sendHomekitUpdate();
        if (mode === winix_api_1.Mode.Manual) {
            return;
        }
        // If we're switching back to auto, the airflow speed will most likely change on the Winix device itself.
        // Pause, get the latest airflow speed, then send the update to Homekit
        this.debug('scheduling homekit update to rotation speed');
        setTimeout(async () => {
            await this.getRotationSpeed(true);
            this.sendHomekitUpdate();
        }, 2000);
    }
    async getRotationSpeed(force = false) {
        if (!force && this.shouldUseCachedValue(this.latestStatus.airflow)) {
            this.debug('getRotationSpeed() (cached)', this.latestStatus.airflow);
            return this.toRotationSpeed(this.latestStatus.airflow);
        }
        let airflow;
        try {
            airflow = await winix_api_1.WinixAPI.getAirflow(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting rotation speed: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.airflow = airflow;
        this.polledWinix();
        this.debug('getRotationSpeed():', airflow);
        return this.toRotationSpeed(airflow);
    }
    async setRotationSpeed(state) {
        const airflow = this.toAirflow(state);
        this.debug(`setRotationSpeed(${state}):`, airflow);
        try {
            await winix_api_1.WinixAPI.setAirflow(this.deviceId, airflow);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error setting rotation speed: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.airflow = airflow;
        this.latestStatus.mode = winix_api_1.Mode.Manual;
        this.sendHomekitUpdate();
    }
    async getAirQuality() {
        if (this.shouldUseCachedValue(this.latestStatus.airQuality)) {
            this.debug('getAirQuality() (cached)', this.latestStatus.airQuality);
            return this.toAirQuality(this.latestStatus.airQuality);
        }
        let airQuality;
        try {
            airQuality = await winix_api_1.WinixAPI.getAirQuality(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting air quality: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.airQuality = airQuality;
        this.polledWinix();
        this.debug('getAirQuality():', airQuality);
        return this.toAirQuality(airQuality);
    }
    async getPlasmawave() {
        if (this.shouldUseCachedValue(this.latestStatus.plasmawave)) {
            this.debug('getPlasmawave() (cached)', this.latestStatus.plasmawave);
            return this.toSwitch(this.latestStatus.plasmawave);
        }
        let plasmawave;
        try {
            plasmawave = await winix_api_1.WinixAPI.getPlasmawave(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting plasmawave state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.plasmawave = plasmawave;
        this.polledWinix();
        this.debug('getPlasmawave():', plasmawave);
        return this.toSwitch(plasmawave);
    }
    async setPlasmawave(state) {
        const plasmawave = this.toPlasmawave(state);
        this.debug(`setPlasmawave(${state}):`, plasmawave);
        try {
            await winix_api_1.WinixAPI.setPlasmawave(this.deviceId, plasmawave);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error setting plasmawave state: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        this.latestStatus.plasmawave = plasmawave;
        this.sendHomekitUpdate();
    }
    async getAmbientLight() {
        if (this.shouldUseCachedValue(this.latestStatus.ambientLight)) {
            this.debug('getAmbientLight() (cached)', this.latestStatus.ambientLight);
            return this.latestStatus.ambientLight;
        }
        let ambientLight;
        try {
            ambientLight = await winix_api_1.WinixAPI.getAmbientLight(this.deviceId);
        }
        catch (e) {
            (0, errors_1.assertError)(e);
            this.error('error getting ambient light: ' + e.message);
            throw new this.platform.HapStatusError(-70402 /* HAPStatus.SERVICE_COMMUNICATION_FAILURE */);
        }
        // Fix ambient light value under 0.0001 warning
        ambientLight = Math.max(ambientLight, MIN_AMBIENT_LIGHT);
        this.latestStatus.ambientLight = ambientLight;
        this.polledWinix();
        this.debug('getAmbientLight():', ambientLight);
        return ambientLight;
    }
    sendHomekitUpdate() {
        this.debug('sendHomekitUpdate()');
        if (this.latestStatus.power !== undefined) {
            this.purifier.updateCharacteristic(this.platform.Characteristic.Active, this.toActiveState(this.latestStatus.power));
            this.purifier.updateCharacteristic(this.platform.Characteristic.CurrentAirPurifierState, this.toCurrentState(this.latestStatus.power));
        }
        if (this.latestStatus.mode !== undefined) {
            this.purifier.updateCharacteristic(this.platform.Characteristic.TargetAirPurifierState, this.toTargetState(this.latestStatus.mode));
        }
        if (this.latestStatus.airflow !== undefined) {
            this.purifier.updateCharacteristic(this.platform.Characteristic.RotationSpeed, this.toRotationSpeed(this.latestStatus.airflow));
        }
        if (this.airQuality !== undefined && this.latestStatus.airQuality !== undefined) {
            this.airQuality?.updateCharacteristic(this.platform.Characteristic.AirQuality, this.toAirQuality(this.latestStatus.airQuality));
        }
        if (this.plasmawave !== undefined && this.latestStatus.plasmawave !== undefined) {
            this.plasmawave?.updateCharacteristic(this.platform.Characteristic.On, this.toSwitch(this.latestStatus.plasmawave));
        }
        if (this.ambientLight !== undefined && this.latestStatus.ambientLight !== undefined) {
            this.ambientLight?.updateCharacteristic(this.platform.Characteristic.CurrentAmbientLightLevel, this.latestStatus.ambientLight);
        }
        if (this.autoSwitch !== undefined && this.latestStatus.mode !== undefined) {
            this.autoSwitch?.updateCharacteristic(this.platform.Characteristic.On, this.toTargetState(this.latestStatus.mode) === this.platform.Characteristic.TargetAirPurifierState.AUTO);
        }
    }
    toActiveState(power) {
        switch (power) {
            case winix_api_1.Power.On:
                return this.platform.Characteristic.Active.ACTIVE;
            case winix_api_1.Power.Off:
                return this.platform.Characteristic.Active.INACTIVE;
        }
    }
    toCurrentState(power) {
        switch (power) {
            case winix_api_1.Power.On:
                return this.platform.Characteristic.CurrentAirPurifierState.PURIFYING_AIR;
            case winix_api_1.Power.Off:
                return this.platform.Characteristic.CurrentAirPurifierState.INACTIVE;
        }
    }
    toTargetState(mode) {
        switch (mode) {
            case winix_api_1.Mode.Auto:
                return this.platform.Characteristic.TargetAirPurifierState.AUTO;
            case winix_api_1.Mode.Manual:
                return this.platform.Characteristic.TargetAirPurifierState.MANUAL;
        }
    }
    toRotationSpeed(airflow) {
        switch (airflow) {
            case winix_api_1.Airflow.Sleep:
                return 0;
            case winix_api_1.Airflow.Low:
                return 25;
            case winix_api_1.Airflow.Medium:
                return 50;
            case winix_api_1.Airflow.High:
                return 75;
            case winix_api_1.Airflow.Turbo:
                return 100;
        }
    }
    toAirflow(state) {
        // Round to nearest 25
        const nearestState = Math.round(state / 25) * 25;
        this.debug(`toAirflow(${state}): ${nearestState}`);
        switch (nearestState) {
            case 0:
                return winix_api_1.Airflow.Sleep;
            case 25:
                return winix_api_1.Airflow.Low;
            case 50:
                return winix_api_1.Airflow.Medium;
            case 75:
                return winix_api_1.Airflow.High;
            case 100:
                return winix_api_1.Airflow.Turbo;
        }
        return nearestState > 100 ? winix_api_1.Airflow.Turbo : winix_api_1.Airflow.Sleep;
    }
    toAirQuality(airQuality) {
        switch (airQuality) {
            case winix_api_1.AirQuality.Good:
                return this.platform.Characteristic.AirQuality.GOOD;
            case winix_api_1.AirQuality.Fair:
                return this.platform.Characteristic.AirQuality.FAIR;
            case winix_api_1.AirQuality.Poor:
                return this.platform.Characteristic.AirQuality.POOR;
            default:
                return this.platform.Characteristic.AirQuality.UNKNOWN;
        }
    }
    shouldUseCachedValue(v) {
        return v !== undefined && Date.now() - this.lastWinixPoll < this.cacheIntervalMs;
    }
    polledWinix() {
        this.lastWinixPoll = Date.now();
    }
    toSwitch(plasmawave) {
        return plasmawave === winix_api_1.Plasmawave.On;
    }
    toPlasmawave(state) {
        return state ? winix_api_1.Plasmawave.On : winix_api_1.Plasmawave.Off;
    }
    getDeviceLogPrefix() {
        return `[${this.platform.logName(this.accessory.context.device)}] `;
    }
    debug(message, ...parameters) {
        this.log.debug(this.getDeviceLogPrefix() + message, ...parameters);
    }
    error(message, ...parameters) {
        this.log.error(this.getDeviceLogPrefix() + message, ...parameters);
    }
}
exports.WinixPurifierAccessory = WinixPurifierAccessory;
